import numpy as np

class CWOFULUserStruct:
    def __init__(self, featureDimension, lambda_, delta_, NoiseScale, Iteration, context_dimension, alpha, budget):
        self.d = featureDimension
        self.A = lambda_ * np.identity(n=self.d)
        self.lambda_ = lambda_
        self.delta_ = delta_
        self.b = np.zeros(self.d)
        self.AInv = np.linalg.inv(self.A)
        self.NoiseScale = NoiseScale
        self.UserTheta = np.zeros(self.d)
        self.iteration = Iteration
        self.context_dimension = context_dimension
        self.alpha = alpha
        self.budget = budget
        self.beta_t = self.NoiseScale * np.sqrt(
            self.d * np.log(1 + self.iteration / (self.d * self.lambda_)) + 2 * np.log(1 / self.delta_)) + 2*np.sqrt(
            self.lambda_) + self.NoiseScale * np.sqrt(self.context_dimension) 

    def updateParameters(self, articlePicked_FeatureVector, click):
        self.w = min(1,2*self.alpha/np.sqrt(np.dot(np.dot(articlePicked_FeatureVector , self.AInv), articlePicked_FeatureVector)))
#        print("CWOFUL self.w",self.w)
        self.A += self.w * np.outer(articlePicked_FeatureVector, articlePicked_FeatureVector)
        self.b += self.w * articlePicked_FeatureVector * click
        self.AInv = np.linalg.inv(self.A)
        self.UserTheta = np.dot(self.AInv, self.b)
#        self.time += 1
        self.beta_t = self.NoiseScale * np.sqrt(
            self.d * np.log(1 + self.iteration / (self.d * self.lambda_)) + 2 * np.log(1 / self.delta_)) + 2*np.sqrt(
            self.lambda_) + self.NoiseScale * np.sqrt(self.context_dimension) 

    def getTheta(self):
        return self.UserTheta

    def getA(self):
        return self.A

    def getProb(self, beta, article_FeatureVector):
        if beta == -1:
            beta = self.beta_t

        mean = np.dot(self.UserTheta, article_FeatureVector)
        var = np.sqrt(np.dot(np.dot(article_FeatureVector, self.AInv), article_FeatureVector))
        pta = mean + beta * var
        return pta

class CWOFUL:
    def __init__(self, dimension, beta, lambda_, delta_, NoiseScale, Iteration, context_dimension, alpha, budget):
        self.users = {}
        self.dimension = dimension
        self.beta = beta
        self.lambda_ = lambda_
        self.delta_ = delta_
        self.NoiseScale = NoiseScale
        self.Iteration = Iteration 
        self.context_dimension = context_dimension
        self.alpha = alpha
        self.budget = budget

        self.CanEstimateUserPreference = True

    def decide(self, pool_articles, userID):
        if userID not in self.users:
            self.users[userID] = CWOFULUserStruct(self.dimension, self.lambda_, self.delta_, self.NoiseScale, self.Iteration, self.context_dimension, self.alpha, self.budget)
        maxPTA = float('-inf')
        articlePicked = None

        for x in pool_articles:
            x_pta = self.users[userID].getProb(self.beta, x.featureVector)
#            print("CWOFUL Mean", self.users[userID].UserTheta)		
#            print("x_pta",x_pta)
            # pick article with highest Prob
            if maxPTA < x_pta:
                articlePicked = x
                maxPTA = x_pta
#        print("CWOFUL articlePicked", articlePicked.id)
        return articlePicked

    def updateParameters(self, articlePicked, click, userID):
        self.users[userID].updateParameters(articlePicked.featureVector, click)

    def getTheta(self, userID):
        return self.users[userID].UserTheta

