from cmath import log, pi, sqrt
import numpy as np
from random import sample, shuffle
import os, sys
import pickle
from StandardEnv import UserManager, ArticleManager
import argparse

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)
from util_functions import gaussianFeature


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = '')
    parser.add_argument('--armnum', type=int, help='Set the number of arms.')
    parser.add_argument('--msave', type=str, help='Set the path for saving the mean value data.')
    parser.add_argument('--asave', type=str, help='Set the path for saving the article data.')
    parser.add_argument('--actionset',type=str, help='Set actionset to be basis vector or uniform from unit sphere.')    
    args = parser.parse_args()
    
    if args.armnum:
        context_dimension = args.armnum
    else:
        context_dimension = 5

    if args.msave:
        meansave = args.msave  
    else:
        meansave = "./Standard/SavedData/MeanData.txt"

    if args.asave:
        articlesave = args.asave  
    else:
        articlesave = "./Standard/SavedData/ArticleData.txt"

    if args.actionset:
        actionset = args.actionset
    else:
        actionset = "basis_vector"  # "basis_vector" or "random"

    ## Environment Settings ##
    NoiseScale = 0.1  # standard deviation of Gaussian noise
    n_users = 1
    n_articles = 5
    poolArticleSize = 5
    targetarmid = poolArticleSize - 1 #target arm for some attack methods, like method from junsun 2018

    if actionset == "basis_vector":
        n_articles = context_dimension  # there can be at most context_dimension number of basis vectors

    ## Set Up Simulation ##
    UM = UserManager(context_dimension, n_users, thetaFunc = gaussianFeature, argv={'l2_limit': 1})
    users = UM.simulateThetafromUsers(actionset)
    while np.argmax(users[0].theta) == targetarmid:
        users = UM.simulateThetafromUsers(actionset)

    print(users[0].theta)
    AM = ArticleManager(context_dimension, n_articles=n_articles)
    articles = AM.simulateArticlePool(actionset)
    
    ## Save Environment Data ##
    fbin = open(meansave,'wb')
    pickle.dump(users,fbin)
    fbin.close
    fbin = open(articlesave,'wb')
    pickle.dump(articles,fbin)
    fbin.close