from collections import Counter
from math import log
import numpy as np 
from random import *

def sigmoid(x):
    return 1/(1+np.exp(-x))

def dsigmoid(x):
    return sigmoid(x)*(1-sigmoid(x))

class FileExists(Exception):
    def __init__(self, filename):
        self.value = filename
    def __str__(self):
        string = """Comment Out! Exception: %s already exists, pass 'force=True' as argument to overwrite file or delete the file."""%(self.value)
        return repr(string)

def getPoolArticleArr(pool_articles):
	article_arr = []
	for x in pool_articles:
		article_arr.append(np.array(x.featureVector))
	return np.array(article_arr)
	
def normalized_gaussianFeature(dimension, argv):
	mean = argv['mean'] if 'mean' in argv else 0.5
	std = argv['std'] if 'std' in argv else 0.1

	mean_vector = np.ones(dimension)*mean
	stdev = np.identity(dimension)*std
	vector = np.random.multivariate_normal(np.zeros(dimension), stdev)

	l2_norm = np.linalg.norm(vector, ord = 2)
	if 'l2_limit' in argv and l2_norm > argv['l2_limit']:
		"This makes it uniform over the circular range"
		vector = (vector / l2_norm)
		vector = vector * (random())
		vector = vector * argv['l2_limit']

	if mean != 0:
		vector = vector + mean_vector

	vectorNormalized = vector
	for i in range(len(vector)):
		vectorNormalized[i] = vector[i]/sum(vector)
	return vectorNormalized
	#return vector

def gaussianFeature(dimension, argv):
	mean = argv['mean'] if 'mean' in argv else 0.5
	std = argv['std'] if 'std' in argv else 0.005

	mean_vector = np.ones(dimension)*mean
	stdev = np.identity(dimension)*std
	vector = np.random.multivariate_normal(np.zeros(dimension), stdev)

	if mean != 0:
		vector = vector + mean_vector    
	
	return vector

def givenfeature(dimension, argv):
	vector = np.zeros(dimension)
#	vector[dimension-1] += 11.5
	return vector

def featureUniform(dimension, argv = None):
	vector = np.array([random() for _ in range(dimension)])

	l2_norm = np.linalg.norm(vector, ord =2)
	
	vector = vector/l2_norm
	return vector

def getBatchStats(arr):
	return np.concatenate((np.array([arr[0]]), np.diff(arr)))

def checkFileExists(filename):
	try:
		with open(filename, 'r'):
			return 1
	except IOError:
		return 0 

def fileOverWriteWarning(filename, force):
	if checkFileExists(filename):
		if force == True:
			print("Warning : fileOverWriteWarning %s"%(filename))
		else:
			raise FileExists(filename)


def vectorize(M):
	# temp = []
	# for i in range(M.shape[0]*M.shape[1]):
	# 	temp.append(M.T.item(i))
	# V = np.asarray(temp)
	# return V
	return np.reshape(M.T, M.shape[0]*M.shape[1])

def matrixize(V, C_dimension):
	# temp = np.zeros(shape = (C_dimension, len(V)/C_dimension))
	# for i in range(len(V)/C_dimension):
	# 	temp.T[i] = V[i*C_dimension : (i+1)*C_dimension]
	# W = temp
	# return W
	#To-do: use numpy built-in function reshape.
	return np.transpose(np.reshape(V, ( int(len(V)/C_dimension), C_dimension)))
